@@ -14,6 +14,7 @@
 #include "update.h"
 #include "main.h"
 #include "global-settings.h"
+#include "utils/misc.h"
 
 using namespace zypp;
 typedef std::set<PoolItem> Candidates;
@@ -22,6 +23,18 @@ extern ZYpp::Ptr God;
 
 static void find_updates( const ResKindSet & kinds, Candidates & candidates, bool all_r );
 
+///////////////////////////////////////////////////////////////////
+/// will go into next libzypp
+namespace zypp_pending
+{
+  template<class TMap>
+  Iterable<typename MapKVIteratorTraits<TMap>::Key_const_iterator> make_map_key_Iterable( const TMap & map_r )
+  {
+    return makeIterable( make_map_key_begin( map_r ), make_map_key_end( map_r ) );
+  }
+} // namespce zypp
+///////////////////////////////////////////////////////////////////
+
 ///////////////////////////////////////////////////////////////////
 namespace
 {
@@ -36,6 +49,117 @@ namespace
     && pi->asKind<Patch>()->restartSuggested();
   }
 
+  /** RNC: Print other-update element */
+  inline std::ostream & xmlPrintOtherUpdateOn( std::ostream & str, const PoolItem & pi_r )
+  {
+    xmlout::Node parent { cout, "update", xmlout::Node::optionalContent, {
+      { "kind", pi_r.kind() },
+      { "name", pi_r.name () },
+      { "edition", pi_r.edition() },
+      { "arch", pi_r.arch() },
+    } };
+    // for packages show also the current installed version (bnc #466599)
+    {
+      const PoolItem & ipi( ui::Selectable::get(pi_r)->installedObj() );
+      if ( ipi )
+      {
+	if ( pi_r.edition() != ipi.edition() )
+	  parent.addAttr( { "edition-old", ipi.edition() } );
+	if ( pi_r.arch() != ipi.arch() )
+	  parent.addAttr( { "arch-old", ipi.arch() } );
+      }
+    }
+
+    dumpAsXmlOn( *parent, pi_r.summary(), "summary" );
+    dumpAsXmlOn( *parent, pi_r.description(), "description" );
+    dumpAsXmlOn( *parent, pi_r.licenseToConfirm(), "license" );
+
+    if ( !pi_r.repoInfo().alias().empty() )
+    {
+      xmlout::Node( *parent, "source", xmlout::Node::optionalContent, {
+	{ "url", pi_r.repoInfo().url().asString() },
+	{ "alias", pi_r.repoInfo().alias() },
+      } );
+    }
+    return str;
+  }
+
+  /** RNC: Print patch-issue-list-element */
+  inline std::ostream & xmlPrintPatchIssueListElementOn( std::ostream & str, Patch::ReferenceIterator begin_r, Patch::ReferenceIterator end_r )
+  {
+    xmlout::Node parent { str, "issue-list", xmlout::Node::optionalContent };
+    for_( it, begin_r, end_r )
+    {
+      xmlout::Node issue { *parent, "issue", xmlout::Node::optionalContent, {
+	{ "type", it.type() },
+	{ "id", it.id() },
+      } };
+      zypp::dumpAsXmlOn( *issue, it.title(), "title" );
+      zypp::dumpAsXmlOn( *issue, it.href(), "href" );
+    }
+    return str;
+  }
+
+  /** RNC: Print patch-update element */
+  inline std::ostream & xmlPrintPatchUpdateOn( std::ostream & str, const PoolItem & pi_r, const PatchHistoryData & patchHistoryData_r )
+  {
+    Patch::constPtr patch = pi_r->asKind<Patch>();
+
+    Patch::InteractiveFlags ignoreFlags = Patch::NoFlags;
+    if ( Zypper::instance().config().reboot_req_non_interactive )
+      ignoreFlags |= Patch::Reboot;
+    if ( LicenseAgreementPolicy::instance()._autoAgreeWithLicenses )
+      ignoreFlags |= Patch::License;
+
+    // write the node
+    xmlout::Node parent { str, "update", xmlout::Node::optionalContent, {
+      { "kind", "patch" },
+      { "name", patch->name () },
+      { "edition", patch->edition() },
+      { "arch", patch->arch() },
+      { "status", textPatchStatus( pi_r ) },
+      { "category", patch->category() },
+      { "severity", patch->severity() },
+      { "pkgmanager", asString( patch->restartSuggested() ) },
+      { "restart", asString( patch->rebootSuggested() ) },
+      { "interactive", asString( patch->interactiveWhenIgnoring(ignoreFlags) ) },
+    } };
+
+    if ( PatchHistoryData::value_type res { patchHistoryData_r[pi_r] }; res != PatchHistoryData::noData )
+    {
+      if ( res.second == pi_r.status().validate() )
+	dumpAsXmlOn( *parent, res.first, "status-since" );
+      else
+	// patch status was changed by a non zypp transaction (not mentioned in the history)
+	DBG << "PatchHistoryData " << res.second << " but " << pi_r << endl;
+    }
+    dumpAsXmlOn( *parent, patch->summary(), "summary" );
+    dumpAsXmlOn( *parent, patch->description(), "description" );
+    dumpAsXmlOn( *parent, patch->licenseToConfirm(), "license" );
+
+    if ( !patch->repoInfo().alias().empty() )
+    {
+      xmlout::Node( *parent, "source", xmlout::Node::optionalContent, {
+	{ "url", patch->repoInfo().url().asString() },
+	{ "alias", patch->repoInfo().alias() },
+      } );
+    }
+
+    dumpAsXmlOn( *parent, patch->timestamp(), "issue-date" );
+    xmlPrintPatchIssueListElementOn( *parent, patch->referencesBegin(), patch->referencesEnd() );
+    return str;
+  }
+
+  /** RNC: Print element containing patch-update-list */
+  template <typename TContainer>
+  inline std::ostream & xmlPrintPatchUpdateListOn( std::ostream & str, std::string nodename_r, TContainer container_r, const PatchHistoryData & patchHistoryData_r )
+  {
+    xmlout::Node parent { str, std::move(nodename_r), xmlout::Node::optionalContent };
+    for ( const auto & pi : container_r )
+      xmlPrintPatchUpdateOn( *parent, pi, patchHistoryData_r );
+    return str;
+  }
+
 } //namespace
 ///////////////////////////////////////////////////////////////////
 
@@ -132,7 +256,7 @@ namespace
 
     struct CategorySort
     {
-      bool operator()( const Patch::Category & lhs, const Patch::Category & rhs )
+      bool operator()( const Patch::Category & lhs, const Patch::Category & rhs ) const
       {
 	if ( lhs == rhs )
 	  return false;
@@ -345,45 +469,8 @@ void patch_check( bool updatestackOnly )
   { zypper.setExitCode( stats.security() ? ZYPPER_EXIT_INF_SEC_UPDATE_NEEDED : ZYPPER_EXIT_INF_UPDATE_NEEDED ); }
 }
 
-static void xml_print_patch( Zypper & zypper, const PoolItem & pi )
-{
-  Patch::constPtr patch = pi->asKind<Patch>();
-
-  cout << " <update ";
-  cout << "name=\"" << patch->name () << "\" ";
-  cout << "edition=\""  << patch->edition() << "\" ";
-  cout << "arch=\""  << patch->arch() << "\" ";
-  cout << "status=\""  << textPatchStatus( pi ) << "\" ";
-  cout << "category=\"" <<  patch->category() << "\" ";
-  cout << "severity=\"" <<  patch->severity() << "\" ";
-  cout << "pkgmanager=\"" << (patch->restartSuggested() ? "true" : "false") << "\" ";
-  cout << "restart=\"" << (patch->rebootSuggested() ? "true" : "false") << "\" ";
-
-  Patch::InteractiveFlags ignoreFlags = Patch::NoFlags;
-  if (zypper.config().reboot_req_non_interactive)
-    ignoreFlags |= Patch::Reboot;
-  if ( LicenseAgreementPolicy::instance()._autoAgreeWithLicenses )
-    ignoreFlags |= Patch::License;
-
-  cout << "interactive=\"" << (patch->interactiveWhenIgnoring(ignoreFlags) ? "true" : "false") << "\" ";
-  cout << "kind=\"patch\"";
-  cout << ">" << endl;
-  cout << "  <summary>" << xml::escape(patch->summary()) << "  </summary>" << endl;
-  cout << "  <description>" << xml::escape(patch->description()) << "</description>" << endl;
-  cout << "  <license>" << xml::escape(patch->licenseToConfirm()) << "</license>" << endl;
-
-  if ( !patch->repoInfo().alias().empty() )
-  {
-    cout << "  <source url=\"" << xml::escape(patch->repoInfo().url().asString());
-    cout << "\" alias=\"" << xml::escape(patch->repoInfo().alias()) << "\"/>" << endl;
-  }
-
-  cout << " </update>" << endl;
-}
-
-
 // returns true if NEEDED! restartSuggested() patches are available
-static bool xml_list_patches (Zypper & zypper, bool all_r )
+static bool xml_list_patches (Zypper & zypper, bool all_r, const PatchHistoryData & patchHistoryData_r )
 {
   const ResPool& pool = God->pool();
 
@@ -409,7 +496,7 @@ static bool xml_list_patches (Zypper & zypper, bool all_r )
       // if updates stack patches are available, show only those
       if ( all_r || !pkg_mgr_available || patchIsNeededRestartSuggested( pi ) )
       {
-	xml_print_patch( zypper, pi );
+	xmlPrintPatchUpdateOn( cout, pi, patchHistoryData_r );
       }
     }
     ++patchcount;
@@ -434,7 +521,7 @@ static bool xml_list_patches (Zypper & zypper, bool all_r )
 	const PoolItem & pi( *it );
 	Patch::constPtr patch = pi->asKind<Patch>();
 	if ( ! patchIsNeededRestartSuggested( pi ) )
-	  xml_print_patch( zypper, pi );
+	  xmlPrintPatchUpdateOn( cout, pi, patchHistoryData_r );
       }
     }
     cout << "</blocked-update-list>" << endl;
@@ -453,47 +540,20 @@ static void xml_list_updates(const ResKindSet & kinds, bool all_r )
 
   for( const PoolItem & pi : candidates )
   {
-    cout << " <update ";
-    cout << "name=\"" << pi.name () << "\" " ;
-    cout << "edition=\""  << pi.edition() << "\" ";
-    cout << "arch=\""  << pi.arch() << "\" ";
-    cout << "kind=\"" << pi.kind() << "\" ";
-    // for packages show also the current installed version (bnc #466599)
-    {
-      const PoolItem & ipi( ui::Selectable::get(pi)->installedObj() );
-      if ( ipi )
-      {
-	if ( pi.edition() != ipi.edition() )
-	  cout << "edition-old=\""  << ipi.edition() << "\" ";
-	if ( pi.arch() != ipi.arch() )
-	  cout << "arch-old=\""  << ipi.arch() << "\" ";
-      }
-    }
-    cout << ">" << endl;
-    cout << "  <summary>" << xml::escape(pi.summary()) << "</summary>" << endl;
-    cout << "  <description>" << xml::escape(pi.description()) << "</description>" << endl;
-    cout << "  <license>" << xml::escape(pi.licenseToConfirm()) << "</license>" << endl;
-
-    if ( !pi.repoInfo().alias().empty() )
-    {
-        cout << "  <source url=\"" << xml::escape(pi.repoInfo().url().asString());
-        cout << "\" alias=\"" << xml::escape(pi.repoInfo().alias()) << "\"/>" << endl;
-    }
-
-    cout << " </update>" << endl;
+    xmlPrintOtherUpdateOn( cout, pi );
   }
 }
 
 // ----------------------------------------------------------------------------
 
 // returns true if NEEDED! restartSuggested() patches are available
-static bool list_patch_updates( Zypper & zypper, bool all_r, const PatchSelector &sel )
+static bool list_patch_updates( Zypper & zypper, bool all_r, const PatchSelector &sel, const PatchHistoryData & historyData_r )
 {
   Table tbl;
-  FillPatchesTable intoTbl( tbl );
+  FillPatchesTable intoTbl( tbl, historyData_r );
 
   Table pmTbl; // only NEEDED! that affect packagemanager (restartSuggested()), they have priority
-  FillPatchesTable intoPMTbl( pmTbl );
+  FillPatchesTable intoPMTbl( pmTbl, historyData_r );
 
   PatchCheckStats stats( zypper.config().exclude_optional_patches );
   CliMatchPatch cliMatchPatch( zypper, sel._requestedPatchDates, sel._requestedPatchCategories, sel._requestedPatchSeverity );
@@ -654,6 +714,8 @@ std::string i18n_kind_updates(const ResKind & kind)
 
 void list_updates(Zypper & zypper, const ResKindSet & kinds, bool best_effort, bool all_r, const PatchSelector &patchSel_r )
 {
+  PatchHistoryData patchHistoryData;	// commonly used by all tables
+
   if (zypper.out().type() == Out::TYPE_XML)
   {
     // TODO: go for XmlNode
@@ -676,7 +738,7 @@ void list_updates(Zypper & zypper, const ResKindSet & kinds, bool best_effort, b
   if(it != localkinds.end())
   {
     if (zypper.out().type() == Out::TYPE_XML)
-      affects_pkgmgr = xml_list_patches( zypper, all_r );
+      affects_pkgmgr = xml_list_patches( zypper, all_r, patchHistoryData );
     else
     {
       if (kinds.size() > 1)
@@ -684,7 +746,7 @@ void list_updates(Zypper & zypper, const ResKindSet & kinds, bool best_effort, b
         zypper.out().info("", Out::NORMAL, Out::TYPE_NORMAL);
         zypper.out().info(i18n_kind_updates(*it), Out::QUIET, Out::TYPE_NORMAL);
       }
-      affects_pkgmgr = list_patch_updates( zypper, all_r, patchSel_r );
+      affects_pkgmgr = list_patch_updates( zypper, all_r, patchSel_r, patchHistoryData );
     }
     localkinds.erase(it);
   }
@@ -745,7 +807,7 @@ void list_updates(Zypper & zypper, const ResKindSet & kinds, bool best_effort, b
     for ( const PoolItem & pi : candidates )
     {
       TableRow tr (cols);
-      tr << "v";
+      tr << computeStatusIndicator( pi );
       if (!hide_repo) {
         tr << pi.repoInfo().asUserString();
       }
@@ -784,24 +846,22 @@ void list_updates(Zypper & zypper, const ResKindSet & kinds, bool best_effort, b
 }
 
 // ----------------------------------------------------------------------------
-
-void list_patches_by_issue( Zypper & zypper, bool all_r, const PatchSelector &sel )
+void list_patches_by_issue( Zypper & zypper, bool all_r, const PatchSelector & sel_r )
 {
-  // --bz, --cve can't be used together with --issue; this case is ruled out
-  // in the initial arguments validation in Zypper.cc
-  Table issueMatchesTbl;
-  FillPatchesTableForIssue intoIssueMatchesTbl( issueMatchesTbl );
+  PatchHistoryData patchHistoryData;	// commonly used by all tables
 
-  const std::set<Issue> &issues = sel._requestedIssues;
-  CliMatchPatch cliMatchPatch( zypper,
-                               sel._requestedPatchDates,
-                               sel._requestedPatchCategories,
-                               sel._requestedPatchSeverity );
+  // Overall CLI filter
   bool only_needed = !all_r;
+  CliMatchPatch cliMatchPatch( zypper,
+                               sel_r._requestedPatchDates,
+                               sel_r._requestedPatchCategories,
+                               sel_r._requestedPatchSeverity );
 
+
+  // pass1 finding PoolItems and their matching issues (pi,itype,iid)
   std::vector<const Issue*> pass2; // on the fly remember anyType issues for pass2
-
-  for ( const Issue & issue : issues )
+  std::map<PoolItem,std::map<std::string,std::set<std::string>>> iresult;
+  for ( const Issue & issue : sel_r._requestedIssues )
   {
     PoolQuery q;
     q.setMatchSubstring();
@@ -816,55 +876,40 @@ void list_patches_by_issue( Zypper & zypper, bool all_r, const PatchSelector &se
       q.addAttribute( sat::SolvAttr::updateReferenceId, issue.id() );
       if ( issue.anyType() && issue.specificId() ) 				// remember for pass2
       {
-	q.addAttribute( sat::SolvAttr::updateReferenceType, issue.id() );	// bnc#941309: let '--issue-bugzilla' also match the type
+	q.addAttribute( sat::SolvAttr::updateReferenceType, issue.id() );	// bnc#941309: let '--issue=bugzilla' also match the type
 	pass2.push_back( &issue );
       }
     }
 
     for_( it, q.begin(), q.end() )
     {
-      PoolItem pi( *it );
-      Patch::constPtr patch = asKind<Patch>(pi);
+      PoolItem pi { *it };
 
       if ( only_needed && ! patchIsApplicable( pi ) )
 	continue;
 
-      if ( ! cliMatchPatch( patch ) )
+      if ( ! cliMatchPatch( pi ) )
       {
-	DBG << patch->ident() << " skipped. (not matching CLI filter)" << endl;
+	DBG << pi.ident() << " skipped. (not matching CLI filter)" << endl;
 	continue;
       }
 
-      // Print details about each match in that solvable:
-      // NOTE: If searching in BOTH updateReferenceId AND updateReferenceType,
-      // we may find matches in both but want to report the issue just once!
-      std::set<std::pair<std::string,std::string>> unifiedResults;
       for_( d, it.matchesBegin(), it.matchesEnd() )
       {
-        std::string itype = d->subFind( sat::SolvAttr::updateReferenceType ).asString();
-
+	std::string itype { d->subFind( sat::SolvAttr::updateReferenceType ).asString() };
 	if ( issue.specificType() && itype != issue.type() )
 	  continue;	// assert correct type of specific IDs
-
 	// remember....
-	unifiedResults.insert( std::make_pair( std::move(itype), d->subFind( sat::SolvAttr::updateReferenceId ).asString() ) );
-      }
-      // print remembered...
-      for ( auto & p : unifiedResults )
-      {
-	intoIssueMatchesTbl( pi, std::move(p.first), std::move(p.second) );
+	iresult[std::move(pi)][std::move(itype)].insert( d->subFind( sat::SolvAttr::updateReferenceId ).asString() );
       }
     }
   }
 
-  // pass2: look for matches in patch summary/description
-  //
-  Table summaryMatchesTbl;
-  FillPatchesTable intoSummaryMatchesTbl( summaryMatchesTbl );
-
-  for ( const Issue* _issue : pass2 )
+  //pass2 (summary/description)
+  std::vector<PoolItem> dresult;	// query will assert unique entries
+  for ( const Issue * _issue : pass2 )
   {
-    const Issue & issue( *_issue );
+    const Issue & issue { *_issue };
     PoolQuery q;
     q.setMatchSubstring();
     q.setCaseSensitive( false );
@@ -874,48 +919,78 @@ void list_patches_by_issue( Zypper & zypper, bool all_r, const PatchSelector &se
 
     for_( it, q.begin(), q.end() )
     {
-      PoolItem pi( *it );
-      Patch::constPtr patch = asKind<Patch>(pi);
+      PoolItem pi { *it };
 
       if ( only_needed && ! patchIsApplicable( pi ) )
 	continue;
 
-      if ( ! cliMatchPatch( patch ) )
+      if ( ! cliMatchPatch( pi ) )
       {
-	DBG << patch->ident() << " skipped. (not matching CLI filter)" << endl;
+	DBG << pi.ident() << " skipped. (not matching CLI filter)" << endl;
 	continue;
       }
 
-      intoSummaryMatchesTbl( pi );
-//	issueMatchesTbl.rows().back().addDetail( str::Str() << "    " << attr.inSolvAttr() << "\t\"" << attr.asString() << "\"" );
-      //! \todo could show a highlighted match with a portion of surrounding
-      //! text. Needs case-insensitive find.
+      if ( ! iresult.count( pi ) )
+      { dresult.push_back( pi ); }
     }
   }
 
+  ///////////////////////////////////////////////////////////////////
   // print result
-  if ( issueMatchesTbl.empty() && summaryMatchesTbl.empty() )
-  {  zypper.out().info(_("No matching issues found.")); }
+  if (zypper.out().type() == Out::TYPE_XML)
+  {
+    xmlout::Node parent { cout, "list-patches-byissue", xmlout::Node::optionalContent };
+    xmlPrintPatchUpdateListOn( *parent, "issue-matches", zypp_pending::make_map_key_Iterable( iresult ), patchHistoryData );
+    xmlPrintPatchUpdateListOn( *parent, "description-matches", dresult, patchHistoryData );
+  }
   else
   {
-    if ( !issueMatchesTbl.empty() )
+    // iresult to table
+    Table issueMatchesTbl;
+    FillPatchesTableForIssue intoIssueMatchesTbl( issueMatchesTbl, patchHistoryData );
+    for ( const auto & res : iresult )
     {
-      if ( !pass2.empty() )
+      const PoolItem & pi { res.first };
+      for ( const auto & ires : res.second )
       {
-        cout << endl;
-        zypper.out().info(_("The following matches in issue numbers have been found:"));
+	const std::string & itype { ires.first };
+	for ( const std::string & iid : ires.second )
+	{ intoIssueMatchesTbl( pi, itype, iid ); }
       }
-      issueMatchesTbl.sort(); // use default sort
-      cout << endl << issueMatchesTbl;
     }
 
-    if ( !summaryMatchesTbl.empty() )
+    // dresult to table
+    Table descrMatchesTbl;
+    FillPatchesTable intoDescrMatchesTbl( descrMatchesTbl, patchHistoryData );
+    for ( const auto & pi : dresult )
+    {
+      intoDescrMatchesTbl( pi );
+    }
+
+
+    if ( issueMatchesTbl.empty() && descrMatchesTbl.empty() )
+    {  zypper.out().info(_("No matching issues found.")); }
+    else
     {
       if ( !issueMatchesTbl.empty() )
-      { cout << endl; }
-      zypper.out().info(_( "Matches in patch descriptions of the following patches have been found:"));
-      summaryMatchesTbl.sort(); // use default sort
-      cout << endl << summaryMatchesTbl;
+      {
+	zypper.out().gap();
+	zypper.out().info(_("The following matches in issue numbers have been found:"));
+
+	issueMatchesTbl.sort(); // use default sort
+	zypper.out().gap();
+	cout << issueMatchesTbl;
+      }
+
+      if ( !descrMatchesTbl.empty() )
+      {
+	zypper.out().gap();
+	zypper.out().info(_( "Matches in patch descriptions of the following patches have been found:"));
+
+	descrMatchesTbl.sort(); // use default sort
+	zypper.out().gap();
+	cout << descrMatchesTbl;
+      }
     }
   }
 }
